@@ -50,6 +50,8 @@
 #include "closestream.h"
 #include "optutils.h"
 #include "fileutils.h"
+#include "loopdev.h"
+#include "buffer.h"
 
 #include "lsblk.h"
 
@@ -63,58 +65,70 @@ static int column_id_to_number(int id);
 
 /* column IDs */
 enum {
-	COL_NAME = 0,
-	COL_KNAME,
-	COL_PATH,
-	COL_MAJMIN,
+	COL_ALIOFF = 0,
+	COL_DALIGN,
+	COL_DAX,
+	COL_DGRAN,
+	COL_DMAX,
+	COL_DZERO,
 	COL_FSAVAIL,
+	COL_FSROOTS,
 	COL_FSSIZE,
 	COL_FSTYPE,
 	COL_FSUSED,
 	COL_FSUSEPERC,
-	COL_TARGET,
-	COL_LABEL,
-	COL_UUID,
-	COL_PTUUID,
-	COL_PTTYPE,
-	COL_PARTTYPE,
-	COL_PARTLABEL,
-	COL_PARTUUID,
-	COL_PARTFLAGS,
-	COL_RA,
-	COL_RO,
-	COL_RM,
+	COL_FSVERSION,
+	COL_GROUP,
+	COL_HCTL,
 	COL_HOTPLUG,
+	COL_KNAME,
+	COL_LABEL,
+	COL_LOGSEC,
+	COL_MAJMIN,
+	COL_MINIO,
+	COL_MODE,
 	COL_MODEL,
+	COL_NAME,
+	COL_OPTIO,
+	COL_OWNER,
+	COL_PARTFLAGS,
+	COL_PARTLABEL,
+	COL_PARTTYPE,
+	COL_PARTTYPENAME,
+	COL_PARTUUID,
+	COL_PATH,
+	COL_PHYSEC,
+	COL_PKNAME,
+	COL_PTTYPE,
+	COL_PTUUID,
+	COL_RA,
+	COL_RAND,
+	COL_REV,
+	COL_RM,
+	COL_RO,
+	COL_ROTA,
+	COL_RQ_SIZE,
+	COL_SCHED,
 	COL_SERIAL,
 	COL_SIZE,
+	COL_START,
 	COL_STATE,
-	COL_OWNER,
-	COL_GROUP,
-	COL_MODE,
-	COL_ALIOFF,
-	COL_MINIO,
-	COL_OPTIO,
-	COL_PHYSEC,
-	COL_LOGSEC,
-	COL_ROTA,
-	COL_SCHED,
-	COL_RQ_SIZE,
+	COL_SUBSYS,
+	COL_TARGET,
+	COL_TARGETS,
+	COL_TRANSPORT,
 	COL_TYPE,
-	COL_DALIGN,
-	COL_DGRAN,
-	COL_DMAX,
-	COL_DZERO,
+	COL_UUID,
+	COL_VENDOR,
 	COL_WSAME,
 	COL_WWN,
-	COL_RAND,
-	COL_PKNAME,
-	COL_HCTL,
-	COL_TRANSPORT,
-	COL_SUBSYS,
-	COL_REV,
-	COL_VENDOR,
 	COL_ZONED,
+	COL_ZONE_SZ,
+	COL_ZONE_WGRAN,
+	COL_ZONE_APP,
+	COL_ZONE_NR,
+	COL_ZONE_OMAX,
+	COL_ZONE_AMAX,
 };
 
 /* basic table settings */
@@ -125,6 +139,7 @@ enum {
 	LSBLK_EXPORT =		(1 << 3),
 	LSBLK_TREE =		(1 << 4),
 	LSBLK_JSON =		(1 << 5),
+	LSBLK_SHELLVAR =	(1 << 6)
 };
 
 /* Types used for qsort() and JSON */
@@ -141,69 +156,76 @@ struct colinfo {
 	const char	*name;		/* header */
 	double		whint;		/* width hint (N < 1 is in percent of termwidth) */
 	int		flags;		/* SCOLS_FL_* */
-	const char      *help;
+	const char	*help;
 	int		type;		/* COLTYPE_* */
 };
 
 /* columns descriptions */
 static struct colinfo infos[] = {
-	[COL_NAME]   = { "NAME",    0.25, SCOLS_FL_NOEXTREMES, N_("device name") },
-	[COL_KNAME]  = { "KNAME",   0.3, 0, N_("internal kernel device name") },
-	[COL_PKNAME] = { "PKNAME",  0.3, 0, N_("internal parent kernel device name") },
-	[COL_PATH]   = { "PATH",    0.3,  0, N_("path to the device node") },
-	[COL_MAJMIN] = { "MAJ:MIN", 6, 0, N_("major:minor device number"), COLTYPE_SORTNUM },
-
-	[COL_FSAVAIL]   = { "FSAVAIL", 5, SCOLS_FL_RIGHT, N_("filesystem size available") },
-	[COL_FSSIZE]    = { "FSSIZE", 5, SCOLS_FL_RIGHT, N_("filesystem size") },
-	[COL_FSTYPE]    = { "FSTYPE", 0.1, SCOLS_FL_TRUNC, N_("filesystem type") },
-	[COL_FSUSED]    = { "FSUSED", 5, SCOLS_FL_RIGHT, N_("filesystem size used") },
-	[COL_FSUSEPERC] = { "FSUSE%", 3, SCOLS_FL_RIGHT, N_("filesystem use percentage") },
-
-	[COL_TARGET] = { "MOUNTPOINT", 0.10, SCOLS_FL_TRUNC, N_("where the device is mounted") },
-	[COL_LABEL]  = { "LABEL",   0.1, 0, N_("filesystem LABEL") },
-	[COL_UUID]   = { "UUID",    36,  0, N_("filesystem UUID") },
-
-	[COL_PTUUID] = { "PTUUID",  36,  0, N_("partition table identifier (usually UUID)") },
-	[COL_PTTYPE] = { "PTTYPE",  0.1, 0, N_("partition table type") },
-
-	[COL_PARTTYPE]  = { "PARTTYPE",  36,  0, N_("partition type UUID") },
-	[COL_PARTLABEL] = { "PARTLABEL", 0.1, 0, N_("partition LABEL") },
-	[COL_PARTUUID]  = { "PARTUUID",  36,  0, N_("partition UUID") },
-	[COL_PARTFLAGS] = { "PARTFLAGS",  36,  0, N_("partition flags") },
-
-	[COL_RA]     = { "RA",      3, SCOLS_FL_RIGHT, N_("read-ahead of the device"), COLTYPE_NUM },
-	[COL_RO]     = { "RO",      1, SCOLS_FL_RIGHT, N_("read-only device"), COLTYPE_BOOL },
-	[COL_RM]     = { "RM",      1, SCOLS_FL_RIGHT, N_("removable device"), COLTYPE_BOOL },
-	[COL_HOTPLUG]= { "HOTPLUG", 1, SCOLS_FL_RIGHT, N_("removable or hotplug device (usb, pcmcia, ...)"), COLTYPE_BOOL },
-	[COL_ROTA]   = { "ROTA",    1, SCOLS_FL_RIGHT, N_("rotational device"), COLTYPE_BOOL },
-	[COL_RAND]   = { "RAND",    1, SCOLS_FL_RIGHT, N_("adds randomness"), COLTYPE_BOOL },
-	[COL_MODEL]  = { "MODEL",   0.1, SCOLS_FL_TRUNC, N_("device identifier") },
-	[COL_SERIAL] = { "SERIAL",  0.1, SCOLS_FL_TRUNC, N_("disk serial number") },
-	[COL_SIZE]   = { "SIZE",    5, SCOLS_FL_RIGHT, N_("size of the device"), COLTYPE_SIZE },
-	[COL_STATE]  = { "STATE",   7, SCOLS_FL_TRUNC, N_("state of the device") },
-	[COL_OWNER]  = { "OWNER",   0.1, SCOLS_FL_TRUNC, N_("user name"), },
-	[COL_GROUP]  = { "GROUP",   0.1, SCOLS_FL_TRUNC, N_("group name") },
-	[COL_MODE]   = { "MODE",    10,   0, N_("device node permissions") },
 	[COL_ALIOFF] = { "ALIGNMENT", 6, SCOLS_FL_RIGHT, N_("alignment offset"), COLTYPE_NUM },
-	[COL_MINIO]  = { "MIN-IO",  6, SCOLS_FL_RIGHT, N_("minimum I/O size"), COLTYPE_NUM },
-	[COL_OPTIO]  = { "OPT-IO",  6, SCOLS_FL_RIGHT, N_("optimal I/O size"), COLTYPE_NUM },
-	[COL_PHYSEC] = { "PHY-SEC", 7, SCOLS_FL_RIGHT, N_("physical sector size"), COLTYPE_NUM },
-	[COL_LOGSEC] = { "LOG-SEC", 7, SCOLS_FL_RIGHT, N_("logical sector size"), COLTYPE_NUM },
-	[COL_SCHED]  = { "SCHED",   0.1, 0, N_("I/O scheduler name") },
-	[COL_RQ_SIZE]= { "RQ-SIZE", 5, SCOLS_FL_RIGHT, N_("request queue size"), COLTYPE_NUM },
-	[COL_TYPE]   = { "TYPE",    4, 0, N_("device type") },
 	[COL_DALIGN] = { "DISC-ALN", 6, SCOLS_FL_RIGHT, N_("discard alignment offset"), COLTYPE_NUM },
-	[COL_DGRAN]  = { "DISC-GRAN", 6, SCOLS_FL_RIGHT, N_("discard granularity"), COLTYPE_SIZE },
-	[COL_DMAX]   = { "DISC-MAX", 6, SCOLS_FL_RIGHT, N_("discard max bytes"), COLTYPE_SIZE },
-	[COL_DZERO]  = { "DISC-ZERO", 1, SCOLS_FL_RIGHT, N_("discard zeroes data"), COLTYPE_BOOL },
-	[COL_WSAME]  = { "WSAME",   6, SCOLS_FL_RIGHT, N_("write same max bytes"), COLTYPE_SIZE },
-	[COL_WWN]    = { "WWN",     18, 0, N_("unique storage identifier") },
-	[COL_HCTL]   = { "HCTL", 10, 0, N_("Host:Channel:Target:Lun for SCSI") },
-	[COL_TRANSPORT] = { "TRAN", 6, 0, N_("device transport type") },
+	[COL_DAX] = { "DAX", 1, SCOLS_FL_RIGHT, N_("dax-capable device"), COLTYPE_BOOL },
+	[COL_DGRAN] = { "DISC-GRAN", 6, SCOLS_FL_RIGHT, N_("discard granularity"), COLTYPE_SIZE },
+	[COL_DMAX] = { "DISC-MAX", 6, SCOLS_FL_RIGHT, N_("discard max bytes"), COLTYPE_SIZE },
+	[COL_DZERO] = { "DISC-ZERO", 1, SCOLS_FL_RIGHT, N_("discard zeroes data"), COLTYPE_BOOL },
+	[COL_FSAVAIL] = { "FSAVAIL", 5, SCOLS_FL_RIGHT, N_("filesystem size available") },
+	[COL_FSROOTS] = { "FSROOTS", 0.1, SCOLS_FL_WRAP, N_("mounted filesystem roots") },
+	[COL_FSSIZE] = { "FSSIZE", 5, SCOLS_FL_RIGHT, N_("filesystem size") },
+	[COL_FSTYPE] = { "FSTYPE", 0.1, SCOLS_FL_TRUNC, N_("filesystem type") },
+	[COL_FSUSED] = { "FSUSED", 5, SCOLS_FL_RIGHT, N_("filesystem size used") },
+	[COL_FSUSEPERC] = { "FSUSE%", 3, SCOLS_FL_RIGHT, N_("filesystem use percentage") },
+	[COL_FSVERSION] = { "FSVER", 0.1, SCOLS_FL_TRUNC, N_("filesystem version") },
+	[COL_GROUP] = { "GROUP", 0.1, SCOLS_FL_TRUNC, N_("group name") },
+	[COL_HCTL] = { "HCTL", 10, 0, N_("Host:Channel:Target:Lun for SCSI") },
+	[COL_HOTPLUG] = { "HOTPLUG", 1, SCOLS_FL_RIGHT, N_("removable or hotplug device (usb, pcmcia, ...)"), COLTYPE_BOOL },
+	[COL_KNAME] = { "KNAME", 0.3, 0, N_("internal kernel device name") },
+	[COL_LABEL] = { "LABEL", 0.1, 0, N_("filesystem LABEL") },
+	[COL_LOGSEC] = { "LOG-SEC", 7, SCOLS_FL_RIGHT, N_("logical sector size"), COLTYPE_NUM },
+	[COL_MAJMIN] = { "MAJ:MIN", 6, 0, N_("major:minor device number"), COLTYPE_SORTNUM },
+	[COL_MINIO] = { "MIN-IO", 6, SCOLS_FL_RIGHT, N_("minimum I/O size"), COLTYPE_NUM },
+	[COL_MODEL] = { "MODEL", 0.1, SCOLS_FL_TRUNC, N_("device identifier") },
+	[COL_MODE] = { "MODE", 10, 0, N_("device node permissions") },
+	[COL_NAME] = { "NAME", 0.25, SCOLS_FL_NOEXTREMES, N_("device name") },
+	[COL_OPTIO] = { "OPT-IO", 6, SCOLS_FL_RIGHT, N_("optimal I/O size"), COLTYPE_NUM },
+	[COL_OWNER] = { "OWNER", 0.1, SCOLS_FL_TRUNC, N_("user name"), },
+	[COL_PARTFLAGS] = { "PARTFLAGS", 36,  0, N_("partition flags") },
+	[COL_PARTLABEL] = { "PARTLABEL", 0.1, 0, N_("partition LABEL") },
+	[COL_PARTTYPENAME]  = { "PARTTYPENAME",  0.1,  0, N_("partition type name") },
+	[COL_PARTTYPE] = { "PARTTYPE", 36,  0, N_("partition type code or UUID") },
+	[COL_PARTUUID] = { "PARTUUID", 36,  0, N_("partition UUID") },
+	[COL_PATH] = { "PATH", 0.3, 0, N_("path to the device node") },
+	[COL_PHYSEC] = { "PHY-SEC", 7, SCOLS_FL_RIGHT, N_("physical sector size"), COLTYPE_NUM },
+	[COL_PKNAME] = { "PKNAME", 0.3, 0, N_("internal parent kernel device name") },
+	[COL_PTTYPE] = { "PTTYPE", 0.1, 0, N_("partition table type") },
+	[COL_PTUUID] = { "PTUUID", 36,  0, N_("partition table identifier (usually UUID)") },
+	[COL_RAND] = { "RAND", 1, SCOLS_FL_RIGHT, N_("adds randomness"), COLTYPE_BOOL },
+	[COL_RA] = { "RA", 3, SCOLS_FL_RIGHT, N_("read-ahead of the device"), COLTYPE_NUM },
+	[COL_REV] = { "REV", 4, SCOLS_FL_RIGHT, N_("device revision") },
+	[COL_RM] = { "RM", 1, SCOLS_FL_RIGHT, N_("removable device"), COLTYPE_BOOL },
+	[COL_ROTA] = { "ROTA", 1, SCOLS_FL_RIGHT, N_("rotational device"), COLTYPE_BOOL },
+	[COL_RO] = { "RO", 1, SCOLS_FL_RIGHT, N_("read-only device"), COLTYPE_BOOL },
+	[COL_RQ_SIZE]= { "RQ-SIZE", 5, SCOLS_FL_RIGHT, N_("request queue size"), COLTYPE_NUM },
+	[COL_SCHED] = { "SCHED", 0.1, 0, N_("I/O scheduler name") },
+	[COL_SERIAL] = { "SERIAL", 0.1, SCOLS_FL_TRUNC, N_("disk serial number") },
+	[COL_SIZE] = { "SIZE", 5, SCOLS_FL_RIGHT, N_("size of the device"), COLTYPE_SIZE },
+	[COL_START] = { "START", 5, SCOLS_FL_RIGHT, N_("partition start offset"), COLTYPE_NUM },
+	[COL_STATE] = { "STATE", 7, SCOLS_FL_TRUNC, N_("state of the device") },
 	[COL_SUBSYS] = { "SUBSYSTEMS", 0.1, SCOLS_FL_NOEXTREMES, N_("de-duplicated chain of subsystems") },
-	[COL_REV]    = { "REV",   4, SCOLS_FL_RIGHT, N_("device revision") },
+	[COL_TARGETS] = { "MOUNTPOINTS", 0.10, SCOLS_FL_WRAP,  N_("all locations where device is mounted") },
+	[COL_TARGET] = { "MOUNTPOINT", 0.10, SCOLS_FL_TRUNC, N_("where the device is mounted") },
+	[COL_TRANSPORT] = { "TRAN", 6, 0, N_("device transport type") },
+	[COL_TYPE] = { "TYPE", 4, 0, N_("device type") },
+	[COL_UUID] = { "UUID", 36,  0, N_("filesystem UUID") },
 	[COL_VENDOR] = { "VENDOR", 0.1, SCOLS_FL_TRUNC, N_("device vendor") },
-	[COL_ZONED]  = { "ZONED", 0.3, 0, N_("zone model") },
+	[COL_WSAME] = { "WSAME", 6, SCOLS_FL_RIGHT, N_("write same max bytes"), COLTYPE_SIZE },
+	[COL_WWN] = { "WWN", 18, 0, N_("unique storage identifier") },
+	[COL_ZONED] = { "ZONED", 0.3, 0, N_("zone model") },
+	[COL_ZONE_SZ] = { "ZONE-SZ", 9, SCOLS_FL_RIGHT, N_("zone size"), COLTYPE_NUM },
+	[COL_ZONE_WGRAN] = { "ZONE-WGRAN", 10, SCOLS_FL_RIGHT, N_("zone write granularity"), COLTYPE_NUM },
+	[COL_ZONE_APP] = { "ZONE-APP", 11, SCOLS_FL_RIGHT, N_("zone append max bytes"), COLTYPE_NUM },
+	[COL_ZONE_NR] = { "ZONE-NR", 8, SCOLS_FL_RIGHT, N_("number of zones"), COLTYPE_NUM },
+	[COL_ZONE_OMAX] = { "ZONE-OMAX", 10, SCOLS_FL_RIGHT, N_("maximum number of open zones"), COLTYPE_NUM },
+	[COL_ZONE_AMAX] = { "ZONE-AMAX", 10, SCOLS_FL_RIGHT, N_("maximum number of active zones"), COLTYPE_NUM },
 };
 
 struct lsblk *lsblk;	/* global handler */
@@ -613,10 +635,9 @@ static char *get_vfs_attribute(struct lsblk_device *dev, int id)
 {
 	char *sizestr;
 	uint64_t vfs_attr = 0;
-	char *mnt;
 
 	if (!dev->fsstat.f_blocks) {
-		mnt = lsblk_device_get_mountpoint(dev);
+		const char *mnt = lsblk_device_get_mountpoint(dev);
 		if (!mnt || dev->is_swap)
 			return NULL;
 		if (statvfs(mnt, &dev->fsstat) != 0)
@@ -700,6 +721,25 @@ static uint64_t device_get_discard_granularity(struct lsblk_device *dev)
 	return dev->discard_granularity;
 }
 
+static void device_read_bytes(struct lsblk_device *dev, char *path, char **str,
+			      uint64_t *sortdata)
+{
+	uint64_t x;
+
+	if (lsblk->bytes) {
+		ul_path_read_string(dev->sysfs, str, path);
+		if (sortdata)
+			str2u64(*str, sortdata);
+		return;
+	}
+
+	if (ul_path_read_u64(dev->sysfs, &x, path) == 0) {
+		*str = size_to_human_string(SIZE_SUFFIX_1LETTER, x);
+		if (sortdata)
+			*sortdata = x;
+	}
+}
+
 /*
  * Generates data (string) for column specified by column ID for specified device. If sortdata
  * is not NULL then returns number usable to sort the column if the data are available for the
@@ -711,7 +751,7 @@ static char *device_get_data(
 		int id,					/* column ID (COL_*) */
 		uint64_t *sortdata)			/* returns sort data as number */
 {
-	struct lsblk_devprop *prop;
+	struct lsblk_devprop *prop = NULL;
 	char *str = NULL;
 
 	switch(id) {
@@ -730,30 +770,42 @@ static char *device_get_data(
 			str = xstrdup(dev->filename);
 		break;
 	case COL_OWNER:
-	{
-		struct stat *st = device_get_stat(dev);
-		struct passwd *pw = st ? getpwuid(st->st_uid) : NULL;
-		if (pw)
-			str = xstrdup(pw->pw_name);
+		if (lsblk->sysroot)
+			prop = lsblk_device_get_properties(dev);
+		if (prop && prop->owner) {
+			str = xstrdup(prop->owner);
+		} else {
+			struct stat *st = device_get_stat(dev);
+			struct passwd *pw = st ? getpwuid(st->st_uid) : NULL;
+			if (pw)
+				str = xstrdup(pw->pw_name);
+		}
 		break;
-	}
 	case COL_GROUP:
-	{
-		struct stat *st = device_get_stat(dev);
-		struct group *gr = st ? getgrgid(st->st_gid) : NULL;
-		if (gr)
-			str = xstrdup(gr->gr_name);
+		if (lsblk->sysroot)
+			prop = lsblk_device_get_properties(dev);
+		if (prop && prop->group) {
+			str = xstrdup(prop->group);
+		} else {
+			struct stat *st = device_get_stat(dev);
+			struct group *gr = st ? getgrgid(st->st_gid) : NULL;
+			if (gr)
+				str = xstrdup(gr->gr_name);
+		}
 		break;
-	}
 	case COL_MODE:
-	{
-		struct stat *st = device_get_stat(dev);
-		char md[11] = { '\0' };
+		if (lsblk->sysroot)
+			prop = lsblk_device_get_properties(dev);
+		if (prop && prop->mode) {
+			str = xstrdup(prop->mode);
+		} else {
+			struct stat *st = device_get_stat(dev);
+			char md[11] = { '\0' };
 
-		if (st)
-			str = xstrdup(xstrmode(st->st_mode, md));
+			if (st)
+				str = xstrdup(xstrmode(st->st_mode, md));
+		}
 		break;
-	}
 	case COL_MAJMIN:
 		if (is_parsable(lsblk))
 			xasprintf(&str, "%u:%u", dev->maj, dev->min);
@@ -773,9 +825,54 @@ static char *device_get_data(
 	case COL_FSUSEPERC:
 		str = get_vfs_attribute(dev, id);
 		break;
-	case COL_TARGET:
-		str = xstrdup(lsblk_device_get_mountpoint(dev));
+	case COL_FSVERSION:
+		prop = lsblk_device_get_properties(dev);
+		if (prop && prop->fsversion)
+			str = xstrdup(prop->fsversion);
 		break;
+	case COL_TARGET:
+	{
+		const char *p = lsblk_device_get_mountpoint(dev);
+		if (p)
+			str = xstrdup(p);
+		break;
+	}
+	case COL_TARGETS:
+	{
+		size_t i, n = 0;
+		struct ul_buffer buf = UL_INIT_BUFFER;
+		struct libmnt_fs **fss = lsblk_device_get_filesystems(dev, &n);
+
+		for (i = 0; i < n; i++) {
+			struct libmnt_fs *fs = fss[i];
+			if (mnt_fs_is_swaparea(fs))
+				ul_buffer_append_string(&buf, "[SWAP]");
+			else
+				ul_buffer_append_string(&buf, mnt_fs_get_target(fs));
+			if (i + 1 < n)
+				ul_buffer_append_data(&buf, "\n", 1);
+		}
+		str = ul_buffer_get_data(&buf, NULL, NULL);
+		break;
+	}
+	case COL_FSROOTS:
+	{
+		size_t i, n = 0;
+		struct ul_buffer buf = UL_INIT_BUFFER;
+		struct libmnt_fs **fss = lsblk_device_get_filesystems(dev, &n);
+
+		for (i = 0; i < n; i++) {
+			struct libmnt_fs *fs = fss[i];
+			const char *root = mnt_fs_get_root(fs);
+			if (mnt_fs_is_swaparea(fs))
+				continue;
+			ul_buffer_append_string(&buf, root ? root : "/");
+			if (i + 1 < n)
+				ul_buffer_append_data(&buf, "\n", 1);
+		}
+		str = ul_buffer_get_data(&buf, NULL, NULL);
+		break;
+	}
 	case COL_LABEL:
 		prop = lsblk_device_get_properties(dev);
 		if (prop && prop->label)
@@ -801,6 +898,15 @@ static char *device_get_data(
 		if (prop && prop->parttype)
 			str = xstrdup(prop->parttype);
 		break;
+	case COL_PARTTYPENAME:
+		prop = lsblk_device_get_properties(dev);
+		if (prop && prop->parttype && prop->pttype) {
+			const char *x = lsblk_parttype_code_to_string(
+						prop->parttype, prop->pttype);
+			if (x)
+				str = xstrdup(x);
+		}
+		break;
 	case COL_PARTLABEL:
 		prop = lsblk_device_get_properties(dev);
 		if (prop && prop->partlabel)
@@ -868,8 +974,6 @@ static char *device_get_data(
 			ul_path_read_string(dev->sysfs, &str, "device/vendor");
 		break;
 	case COL_SIZE:
-		if (!dev->size)
-			break;
 		if (lsblk->bytes)
 			xasprintf(&str, "%ju", dev->size);
 		else
@@ -877,6 +981,11 @@ static char *device_get_data(
 		if (sortdata)
 			*sortdata = dev->size;
 		break;
+	case COL_START:
+		ul_path_read_string(dev->sysfs, &str, "start");
+		if (sortdata)
+			str2u64(str, sortdata);
+		break;
 	case COL_STATE:
 		if (!device_is_partition(dev) && !dev->dm_name)
 			ul_path_read_string(dev->sysfs, &str, "device/state");
@@ -956,18 +1065,7 @@ static char *device_get_data(
 		}
 		break;
 	case COL_DMAX:
-		if (lsblk->bytes) {
-			ul_path_read_string(dev->sysfs, &str, "queue/discard_max_bytes");
-			if (sortdata)
-				str2u64(str, sortdata);
-		} else {
-			uint64_t x;
-			if (ul_path_read_u64(dev->sysfs, &x, "queue/discard_max_bytes") == 0) {
-				str = size_to_human_string(SIZE_SUFFIX_1LETTER, x);
-				if (sortdata)
-					*sortdata = x;
-			}
-		}
+		device_read_bytes(dev, "queue/discard_max_bytes", &str, sortdata);
 		break;
 	case COL_DZERO:
 		if (device_get_discard_granularity(dev) > 0)
@@ -976,25 +1074,56 @@ static char *device_get_data(
 			str = xstrdup("0");
 		break;
 	case COL_WSAME:
-		if (lsblk->bytes) {
-			ul_path_read_string(dev->sysfs, &str, "queue/write_same_max_bytes");
-			if (sortdata)
-				str2u64(str, sortdata);
-		} else {
-			uint64_t x;
-
-			if (ul_path_read_u64(dev->sysfs, &x, "queue/write_same_max_bytes") == 0) {
-				str = size_to_human_string(SIZE_SUFFIX_1LETTER, x);
-				if (sortdata)
-					*sortdata = x;
-			}
-		}
+		device_read_bytes(dev, "queue/write_same_max_bytes", &str, sortdata);
 		if (!str)
 			str = xstrdup("0");
 		break;
 	case COL_ZONED:
 		ul_path_read_string(dev->sysfs, &str, "queue/zoned");
 		break;
+	case COL_ZONE_SZ:
+	{
+		uint64_t x;
+
+		if (ul_path_read_u64(dev->sysfs, &x, "queue/chunk_sectors") == 0) {
+			x <<= 9;
+			if (lsblk->bytes)
+				xasprintf(&str, "%ju", x);
+			else
+				str = size_to_human_string(SIZE_SUFFIX_1LETTER, x);
+			if (sortdata)
+				*sortdata = x;
+		}
+		break;
+	}
+	case COL_ZONE_WGRAN:
+		device_read_bytes(dev, "queue/zone_write_granularity", &str, sortdata);
+		break;
+	case COL_ZONE_APP:
+		device_read_bytes(dev, "queue/zone_append_max_bytes", &str, sortdata);
+		break;
+	case COL_ZONE_NR:
+		ul_path_read_string(dev->sysfs, &str, "queue/nr_zones");
+		if (sortdata)
+			str2u64(str, sortdata);
+		break;
+	case COL_ZONE_OMAX:
+		ul_path_read_string(dev->sysfs, &str, "queue/max_open_zones");
+		if (!str)
+			str = xstrdup("0");
+		if (sortdata)
+			str2u64(str, sortdata);
+		break;
+	case COL_ZONE_AMAX:
+		ul_path_read_string(dev->sysfs, &str, "queue/max_active_zones");
+		if (!str)
+			str = xstrdup("0");
+		if (sortdata)
+			str2u64(str, sortdata);
+		break;
+	case COL_DAX:
+		ul_path_read_string(dev->sysfs, &str, "queue/dax");
+		break;
 	};
 
 	return str;
@@ -1022,8 +1151,8 @@ static void device_to_scols(
 	if (!parent && dev->wholedisk)
 		parent = dev->wholedisk;
 
-	/* Do not print device more than one in --list mode */
-	if (!(lsblk->flags & LSBLK_TREE) && dev->is_printed)
+	/* Do not print device more than once on --list if tree order is not requested */
+	if (!(lsblk->flags & LSBLK_TREE) && !lsblk->force_tree_order && dev->is_printed)
 		return;
 
 	if (lsblk->merge && list_count_entries(&dev->parents) > 1) {
@@ -1110,6 +1239,23 @@ static void devtree_to_scols(struct lsblk_devtree *tr, struct libscols_table *ta
 		device_to_scols(dev, NULL, tab, NULL);
 }
 
+static int ignore_empty(struct lsblk_device *dev)
+{
+	/* show all non-empty devices */
+	if (dev->size)
+		return 0;
+
+	if (lsblk->noempty && dev->size == 0)
+		return 1;
+
+	/* ignore empty loop devices without backing file */
+	if (dev->maj == LOOPDEV_MAJOR &&
+	    !loopdev_has_backing_file(dev->filename))
+		return 1;
+
+	return 0;
+}
+
 /*
  * Reads very basic information about the device from sysfs into the device struct
  */
@@ -1122,6 +1268,11 @@ static int initialize_device(struct lsblk_device *dev,
 	DBG(DEV, ul_debugobj(dev, "initialize %s [wholedisk=%p %s]",
 			name, wholedisk, wholedisk ? wholedisk->name : ""));
 
+	if (sysfs_devname_is_hidden(lsblk->sysroot, name)) {
+		DBG(DEV, ul_debugobj(dev, "%s: hidden, ignore", name));
+		return -1;
+	}
+
 	dev->name = xstrdup(name);
 
 	if (wholedisk) {
@@ -1156,7 +1307,7 @@ static int initialize_device(struct lsblk_device *dev,
 		dev->size <<= 9;					/* in bytes */
 
 	/* Ignore devices of zero size */
-	if (!lsblk->all_devices && dev->size == 0) {
+	if (!lsblk->all_devices && ignore_empty(dev)) {
 		DBG(DEV, ul_debugobj(dev, "zero size device -- ignore"));
 		return -1;
 	}
@@ -1208,6 +1359,26 @@ static struct lsblk_device *devtree_get_device_or_new(struct lsblk_devtree *tr,
 	return dev;
 }
 
+static struct lsblk_device *devtree_pktcdvd_get_dep(
+			struct lsblk_devtree *tr,
+			struct lsblk_device *dev,
+			int want_slave)
+{
+	char buf[PATH_MAX], *name;
+	dev_t devno;
+
+	devno = lsblk_devtree_pktcdvd_get_mate(tr,
+			makedev(dev->maj, dev->min), !want_slave);
+	if (!devno)
+		return NULL;
+
+	name = sysfs_devno_to_devname(devno, buf, sizeof(buf));
+	if (!name)
+		return NULL;
+
+	return devtree_get_device_or_new(tr, NULL, name);
+}
+
 static int process_dependencies(
 			struct lsblk_devtree *tr,
 			struct lsblk_device *dev,
@@ -1298,6 +1469,7 @@ static int process_dependencies(
 	DIR *dir;
 	struct dirent *d;
 	const char *depname;
+	struct lsblk_device *dep = NULL;
 
 	assert(dev);
 
@@ -1312,21 +1484,20 @@ static int process_dependencies(
 
 	if (!(lsblk->inverse ? dev->nslaves : dev->nholders)) {
 		DBG(DEV, ul_debugobj(dev, " ignore (no slaves/holders)"));
-		return 0;
+		goto done;
 	}
 
 	depname = lsblk->inverse ? "slaves" : "holders";
 	dir = ul_path_opendir(dev->sysfs, depname);
 	if (!dir) {
 		DBG(DEV, ul_debugobj(dev, " ignore (no slaves/holders directory)"));
-		return 0;
+		goto done;
 	}
 	ul_path_close_dirfd(dev->sysfs);
 
 	DBG(DEV, ul_debugobj(dev, " %s: checking for '%s' dependence", dev->name, depname));
 
 	while ((d = xreaddir(dir))) {
-		struct lsblk_device *dep = NULL;
 		struct lsblk_device *disk = NULL;
 
 		/* Is the dependency a partition? */
@@ -1377,8 +1548,14 @@ next:
 			ul_path_close_dirfd(disk->sysfs);
 	}
 	closedir(dir);
+done:
+	dep = devtree_pktcdvd_get_dep(tr, dev, lsblk->inverse);
+
+	if (dep && lsblk_device_new_dependence(dev, dep) == 0) {
+		lsblk_devtree_remove_root(tr, dep);
+		process_dependencies(tr, dep, lsblk->inverse ? 0 : 1);
+	}
 
-	DBG(DEV, ul_debugobj(dev, "%s: checking for '%s' -- done", dev->name, depname));
 	return 0;
 }
 
@@ -1722,10 +1899,12 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_("List information about block devices.\n"), out);
 
 	fputs(USAGE_OPTIONS, out);
+	fputs(_(" -A, --noempty        don't print empty devices\n"), out);
 	fputs(_(" -D, --discard        print discard capabilities\n"), out);
 	fputs(_(" -E, --dedup <column> de-duplicate output by <column>\n"), out);
 	fputs(_(" -I, --include <list> show only devices with specified major numbers\n"), out);
 	fputs(_(" -J, --json           use JSON output format\n"), out);
+	fputs(_(" -M, --merge          group parents of sub-trees (usable for RAIDs, Multi-path)\n"), out);
 	fputs(_(" -O, --output-all     output all columns\n"), out);
 	fputs(_(" -P, --pairs          use key=\"value\" output format\n"), out);
 	fputs(_(" -S, --scsi           output info about SCSI devices\n"), out);
@@ -1737,7 +1916,6 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_(" -f, --fs             output info about filesystems\n"), out);
 	fputs(_(" -i, --ascii          use ascii characters only\n"), out);
 	fputs(_(" -l, --list           use list format output\n"), out);
-	fputs(_(" -M, --merge          group parents of sub-trees (usable for RAIDs, Multi-path)\n"), out);
 	fputs(_(" -m, --perms          output info about permissions\n"), out);
 	fputs(_(" -n, --noheadings     don't print headings\n"), out);
 	fputs(_(" -o, --output <list>  output columns\n"), out);
@@ -1745,8 +1923,10 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_(" -r, --raw            use raw output format\n"), out);
 	fputs(_(" -s, --inverse        inverse dependencies\n"), out);
 	fputs(_(" -t, --topology       output info about topology\n"), out);
-	fputs(_(" -z, --zoned          print zone model\n"), out);
+	fputs(_(" -w, --width <num>    specifies output width as number of characters\n"), out);
 	fputs(_(" -x, --sort <column>  sort output by <column>\n"), out);
+	fputs(_(" -y, --shell          use column names to be usable as shell variable identifiers\n"), out);
+	fputs(_(" -z, --zoned          print zone related information\n"), out);
 	fputs(_("     --sysroot <dir>  use specified directory as system root\n"), out);
 	fputs(USAGE_SEPARATOR, out);
 	printf(USAGE_HELP_OPTIONS(22));
@@ -1754,7 +1934,7 @@ static void __attribute__((__noreturn__)) usage(void)
 	fprintf(out, USAGE_COLUMNS);
 
 	for (i = 0; i < ARRAY_SIZE(infos); i++)
-		fprintf(out, " %11s  %s\n", infos[i].name, _(infos[i].help));
+		fprintf(out, " %12s  %s\n", infos[i].name, _(infos[i].help));
 
 	printf(USAGE_MAN_TAIL("lsblk(8)"));
 
@@ -1780,6 +1960,7 @@ int main(int argc, char *argv[])
 	int c, status = EXIT_FAILURE;
 	char *outarg = NULL;
 	size_t i;
+	unsigned int width = 0;
 	int force_tree = 0, has_tree_col = 0;
 
 	enum {
@@ -1790,6 +1971,7 @@ int main(int argc, char *argv[])
 		{ "all",	no_argument,       NULL, 'a' },
 		{ "bytes",      no_argument,       NULL, 'b' },
 		{ "nodeps",     no_argument,       NULL, 'd' },
+		{ "noempty",    no_argument,       NULL, 'A' },
 		{ "discard",    no_argument,       NULL, 'D' },
 		{ "dedup",      required_argument, NULL, 'E' },
 		{ "zoned",      no_argument,       NULL, 'z' },
@@ -1813,8 +1995,10 @@ int main(int argc, char *argv[])
 		{ "scsi",       no_argument,       NULL, 'S' },
 		{ "sort",	required_argument, NULL, 'x' },
 		{ "sysroot",    required_argument, NULL, OPT_SYSROOT },
+		{ "shell",      no_argument,       NULL, 'y' },
 		{ "tree",       optional_argument, NULL, 'T' },
 		{ "version",    no_argument,       NULL, 'V' },
+		{ "width",	required_argument, NULL, 'w' },
 		{ NULL, 0, NULL, 0 },
 	};
 
@@ -1842,11 +2026,15 @@ int main(int argc, char *argv[])
 	lsblk_init_debug();
 
 	while((c = getopt_long(argc, argv,
-			       "abdDzE:e:fhJlnMmo:OpPiI:rstVST:x:", longopts, NULL)) != -1) {
+				"AabdDzE:e:fhJlnMmo:OpPiI:rstVST::w:x:y",
+				longopts, NULL)) != -1) {
 
 		err_exclusive_options(c, longopts, excl, excl_st);
 
 		switch(c) {
+		case 'A':
+			lsblk->noempty = 1;
+			break;
 		case 'a':
 			lsblk->all_devices = 1;
 			break;
@@ -1866,6 +2054,12 @@ int main(int argc, char *argv[])
 		case 'z':
 			add_uniq_column(COL_NAME);
 			add_uniq_column(COL_ZONED);
+			add_uniq_column(COL_ZONE_SZ);
+			add_uniq_column(COL_ZONE_NR);
+			add_uniq_column(COL_ZONE_AMAX);
+			add_uniq_column(COL_ZONE_OMAX);
+			add_uniq_column(COL_ZONE_APP);
+			add_uniq_column(COL_ZONE_WGRAN);
 			break;
 		case 'e':
 			parse_excludes(optarg);
@@ -1896,6 +2090,9 @@ int main(int argc, char *argv[])
 			lsblk->flags |= LSBLK_EXPORT;
 			lsblk->flags &= ~LSBLK_TREE;	/* disable the default */
 			break;
+		case 'y':
+			lsblk->flags |= LSBLK_SHELLVAR;
+			break;
 		case 'i':
 			lsblk->flags |= LSBLK_ASCII;
 			break;
@@ -1912,11 +2109,12 @@ int main(int argc, char *argv[])
 		case 'f':
 			add_uniq_column(COL_NAME);
 			add_uniq_column(COL_FSTYPE);
+			add_uniq_column(COL_FSVERSION);
 			add_uniq_column(COL_LABEL);
 			add_uniq_column(COL_UUID);
 			add_uniq_column(COL_FSAVAIL);
 			add_uniq_column(COL_FSUSEPERC);
-			add_uniq_column(COL_TARGET);
+			add_uniq_column(COL_TARGETS);
 			break;
 		case 'm':
 			add_uniq_column(COL_NAME);
@@ -1947,12 +2145,16 @@ int main(int argc, char *argv[])
 			add_uniq_column(COL_VENDOR);
 			add_uniq_column(COL_MODEL);
 			add_uniq_column(COL_REV);
+			add_uniq_column(COL_SERIAL);
 			add_uniq_column(COL_TRANSPORT);
 			break;
 		case 'T':
 			force_tree = 1;
-			if (optarg)
+			if (optarg) {
+				if (*optarg == '=')
+					optarg++;
 				lsblk->tree_id = column_name_to_id(optarg, strlen(optarg));
+			}
 			break;
 		case OPT_SYSROOT:
 			lsblk->sysroot = optarg;
@@ -1963,6 +2165,9 @@ int main(int argc, char *argv[])
 				break;
 			errtryhelp(EXIT_FAILURE);
 			break;
+		case 'w':
+			width = strtou32_or_err(optarg, _("invalid output width number argument"));
+			break;
 		case 'x':
 			lsblk->flags &= ~LSBLK_TREE; /* disable the default */
 			lsblk->sort_id = column_name_to_id(optarg, strlen(optarg));
@@ -1992,7 +2197,7 @@ int main(int argc, char *argv[])
 		add_column(COL_SIZE);
 		add_column(COL_RO);
 		add_column(COL_TYPE);
-		add_column(COL_TARGET);
+		add_column(COL_TARGETS);
 	}
 
 	if (outarg && string_add_to_idarray(outarg, columns, ARRAY_SIZE(columns),
@@ -2007,8 +2212,9 @@ int main(int argc, char *argv[])
 		 * /sys is no more sorted */
 		lsblk->sort_id = COL_MAJMIN;
 
-	/* For --inverse --list we still follow parent->child relation */
-	if (lsblk->inverse && !(lsblk->flags & LSBLK_TREE))
+	/* For --{inverse,raw,pairs} --list we still follow parent->child relation */
+	if (!(lsblk->flags & LSBLK_TREE)
+	    && (lsblk->inverse || lsblk->flags & LSBLK_EXPORT || lsblk->flags & LSBLK_RAW))
 		lsblk->force_tree_order = 1;
 
 	if (lsblk->sort_id >= 0 && column_id_to_number(lsblk->sort_id) < 0) {
@@ -2034,12 +2240,17 @@ int main(int argc, char *argv[])
 		errx(EXIT_FAILURE, _("failed to allocate output table"));
 	scols_table_enable_raw(lsblk->table, !!(lsblk->flags & LSBLK_RAW));
 	scols_table_enable_export(lsblk->table, !!(lsblk->flags & LSBLK_EXPORT));
+	scols_table_enable_shellvar(lsblk->table, !!(lsblk->flags & LSBLK_SHELLVAR));
 	scols_table_enable_ascii(lsblk->table, !!(lsblk->flags & LSBLK_ASCII));
 	scols_table_enable_json(lsblk->table, !!(lsblk->flags & LSBLK_JSON));
 	scols_table_enable_noheadings(lsblk->table, !!(lsblk->flags & LSBLK_NOHEADINGS));
 
 	if (lsblk->flags & LSBLK_JSON)
 		scols_table_set_name(lsblk->table, "blockdevices");
+	if (width) {
+		scols_table_set_termwidth(lsblk->table, width);
+		scols_table_set_termforce(lsblk->table, SCOLS_TERMFORCE_ALWAYS);
+	}
 
 	for (i = 0; i < ncolumns; i++) {
 		struct colinfo *ci = get_column_info(i);
@@ -2081,6 +2292,15 @@ int main(int argc, char *argv[])
 			        ci->type == COLTYPE_SORTNUM ? cmp_u64_cells : scols_cmpstr_cells,
 				NULL);
 		}
+		/* multi-line cells (now used for MOUNTPOINTS) */
+		if (fl & SCOLS_FL_WRAP) {
+			scols_column_set_wrapfunc(cl,
+						scols_wrapnl_chunksize,
+						scols_wrapnl_nextchunk,
+						NULL);
+			scols_column_set_safechars(cl, "\n");
+		}
+
 		if (lsblk->flags & LSBLK_JSON) {
 			switch (ci->type) {
 			case COLTYPE_SIZE:
@@ -2088,13 +2308,16 @@ int main(int argc, char *argv[])
 					break;
 				/* fallthrough */
 			case COLTYPE_NUM:
-				 scols_column_set_json_type(cl, SCOLS_JSON_NUMBER);
+				scols_column_set_json_type(cl, SCOLS_JSON_NUMBER);
 				break;
 			case COLTYPE_BOOL:
 				scols_column_set_json_type(cl, SCOLS_JSON_BOOLEAN);
 				break;
 			default:
-				scols_column_set_json_type(cl, SCOLS_JSON_STRING);
+				if (fl & SCOLS_FL_WRAP)
+					scols_column_set_json_type(cl, SCOLS_JSON_ARRAY_STRING);
+				else
+					scols_column_set_json_type(cl, SCOLS_JSON_STRING);
 				break;
 			}
 		}
